import json
import os
import re
import tarfile

import tqdm
from bs4 import BeautifulSoup, Comment
from demoji import replace
from markdown import markdown

counter = 0


def clean_body(body):
    # Check if body is present
    if body is None:
        return None

    html = markdown(body)
    soup = BeautifulSoup(html, features='html.parser')

    # Remove everything in the <details> tag (noisy data)
    for s in soup.findAll('details'):
        s.extract()

    # Remove HTML comments
    for s in soup(text=lambda comment: isinstance(comment, Comment)):
        s.extract()

    # Remove emojis
    body = replace(soup.get_text('\n')).strip()

    # Remove all newlines and replace with single space
    body = re.sub(r'\s+', ' ', body, flags=re.MULTILINE)

    # Remove all links
    body = re.sub(r'http(s)?://\S+', ' ', body, flags=re.MULTILINE)

    # Check if body only contains ASCII characters
    if not body.isascii():
        return None

    return body


def clean_title(title):
    if title is None:
        return None

    # Remove emojis
    title = replace(title).strip()

    # Check if title only contains ASCII characters
    if not title.isascii():
        return None

    return title


def read_issue_obj(obj, enable_filter=True):
    global counter

    if enable_filter:
        if 'PR' in obj['node_id'] or ('pull_request' in obj and obj['pull_request'] is not None):
            return None  # skip since it is a pull request

        if 'assignees' not in obj or obj['assignees'] is None:
            counter += 1
            return None  # skip since it has not been assigned

        if len(obj['assignees']) != 1:
            counter += 1
            return None  # skip since it has multiple assignees

    body = clean_body(obj['body'])
    dirty_body = obj['body']
    title = clean_title(obj['title'])
    title_count = 0 if title is None else len(title.split())

    issue = {
        'id': obj['number'],
        'title': title,
        'body': body,
        'dirty_body': dirty_body,
        'word_count': 0 if body is None else len(body.split()) + title_count,
        'word_count_dirty': 0 if body is None else len(dirty_body.split()) + title_count,
        'state': obj['state'],
        'assignee': None if len(obj['assignees']) == 0 else obj['assignees'][0]['login'],
        'created_at': obj['created_at']
    }

    if issue['title'] is None and issue['body'] is None:
        counter += 1
        return None

    return issue


def read_issue(file):
    global counter

    try:
        obj = json.load(file)
        issue = read_issue_obj(obj)
    except json.decoder.JSONDecodeError:
        # Some downloaded issues are a blank JSON file. We suspect that these issues were available at the time of
        # listing, but they have been since deleted and are not available anymore through the GitHub API, therefore
        # we choose to ignore them. The internal issue IDs for these issues were:
        # 111293876, 116791101, 116805010, 116805553, 116805977, 116901067, 117010737, 117065474, 117067419, 117068152,
        # 117069931, 116803071, 116923175, 116989517, 117063475, 117067644
        counter += 1
        return None

    return issue


def save_set(df, id_from, id_to, name, file_prefix: str):
    dir_name = os.path.dirname(file_prefix)
    if not os.path.isdir(dir_name):
        os.makedirs(dir_name)

    df.loc[id_from:id_to].to_csv(file_prefix + f'{name}_{id_from:06d}_{id_to:06d}.csv')


def clean_all(objs, in_file: str):
    global counter

    counter = 0
    tar = tarfile.open(in_file, 'r:gz')

    for member in tqdm.tqdm(tar.getmembers()):
        if member.name.endswith('.json'):
            f = tar.extractfile(member)

            if f is not None:
                issue = read_issue(f)

                if issue is not None:
                    objs.append(issue)

    return counter
