import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import tqdm
from sklearn.metrics import RocCurveDisplay, roc_auc_score
from sklearn.preprocessing import LabelBinarizer

from .classifier import Classifier
from .dataset import Labelling, SplitData, load_df
from .evaluate import predict
from .load import load_model
from .torch_dataset import Dataset


def predict_proba(model: Classifier, test: Dataset, n_classes: int, force_cpu: bool) -> np.ndarray:
    test_output = predict(model, test, -1, force_cpu)
    out = np.ndarray(shape=(len(test.texts), n_classes))

    for i, res in enumerate(tqdm.tqdm(test_output, desc="Predicting outputs for ROC/AUC")):
        for c, pred_class in enumerate(res.top_indices):
            if pred_class >= 0:
                pred_prob = res.top_values[c]
                out[i, pred_class] = pred_prob

    return out


def compute_auc_roc(model: Classifier, test_data: SplitData, n_classes, labelling: Labelling, force_cpu: bool,
                    out_prefix: str):
    # Encode classes in one-hot encoding for one-vs-rest AUC
    lb = LabelBinarizer()
    lb.fit(range(n_classes))
    y_onehot = lb.transform(test_data.labels)
    y_score = predict_proba(model, Dataset(test_data), n_classes, force_cpu)

    # Color map for classes
    colormap = mpl.colormaps['Spectral'].resampled(n_classes)

    fig, ax = plt.subplots(figsize=(14, 14))

    for assignee, class_n in tqdm.tqdm(labelling.labels.items(), desc="Computing ROC curves"):
        if 0 <= class_n < n_classes:
            lb_class = np.flatnonzero(lb.classes_ == class_n)[0]
            y_true = y_onehot[:, lb_class]

            if len(np.flatnonzero(y_true)) > 0:  # if this class is in the test set
                RocCurveDisplay.from_predictions(
                    y_true,
                    y_score[:, lb_class],
                    ax=ax,
                    name=f"{assignee} ({class_n})",
                    color=colormap(class_n)
                )

    plt.axis("square")
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    plt.title("One-vs-Rest ROC curves")
    plt.legend()
    plt.savefig(out_prefix + ".ovr_curves.png")

    fig, ax = plt.subplots(figsize=(7, 7))
    RocCurveDisplay.from_predictions(
        y_onehot.ravel(),
        y_score.ravel(),
        ax=ax,
        name="micro-average OvR",
        color="darkorange",
    )
    plt.axis("square")
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    plt.title("Micro-averaged One-vs-Rest\nReceiver Operating Characteristic")
    ax.get_legend().remove()
    plt.savefig(out_prefix + ".ovr_avg.png")

    micro_roc_auc_ovr = roc_auc_score(
        y_onehot,
        y_score,
        multi_class="ovr",
        average="micro",
    )

    message = f"Micro-averaged One-vs-Rest ROC AUC score:\n{micro_roc_auc_ovr:.4f}"

    with open(out_prefix + ".auc.txt", "w") as f:
        f.write(message)

    print(message)


def build_curve(path: str, force_cpu: bool):
    if not path.endswith('.pt'):
        raise ValueError("path should point to a pytorch model file")

    pretrained, model, classes = load_model(path, None, force_cpu, False)

    if not pretrained:
        raise FileNotFoundError("Trained model is needed to run predict script")

    out_prefix = path[:-3]
    csv_path = out_prefix + '.labels.csv'
    labelling = Labelling.load(csv_path)
    test_data = SplitData.from_df(load_df(f'issues_test_170001_180000.csv'), labelling, classes)

    compute_auc_roc(model, test_data, classes, labelling, force_cpu, out_prefix)
