import os.path
from dataclasses import dataclass
from typing import Optional

import numpy as np
import pandas as pd
from transformers import DistilBertTokenizer
from transformers.tokenization_utils_base import BatchEncoding

DIR: str = os.path.join(os.path.dirname(__file__), '..', '..', 'out', 'csv')
OUT_COLUMN: str = 'assignee'
IN_COLUMNS: list[str] = ['title', 'body']
IN_JOINED_COLUMN: str = 'title_body'
VALIDATION_PERC = 0.1


def prepare_input(df: pd.DataFrame) -> pd.DataFrame:
    df[IN_JOINED_COLUMN] = df[IN_COLUMNS].apply(lambda row: '\n'.join(row.values.astype(str)), axis=1)
    return df.loc[:, [IN_JOINED_COLUMN, OUT_COLUMN]]


def load_df(csv_filename: str) -> pd.DataFrame:
    df: pd.DataFrame = pd.read_csv(os.path.join(DIR, csv_filename))
    df = df.set_index('id', drop=True)
    return prepare_input(df)


def compute_labels(frames: list[pd.DataFrame]) -> tuple[dict[str, int], list[int]]:
    n: int = 0
    labels_dict: dict[str, int] = {}
    num_bounds: list[int] = [0]

    for frame in frames:
        labels: list[str] = frame[OUT_COLUMN] \
            .drop_duplicates(keep='first') \
            .sort_values() \
            .to_list()

        for label in labels:
            if label not in labels_dict:
                labels_dict[label] = n
                n += 1

        num_bounds.append(n)

    return labels_dict, num_bounds


tokenizer = DistilBertTokenizer.from_pretrained('distilbert-base-uncased', return_token_type_ids=False)


@dataclass
class Labelling:
    labels: dict[str, int]
    labels_rev: dict[int, str]

    def __init__(self, labels: dict[str, int]):
        self.labels = labels
        self.labels_rev = {v: k for k, v in self.labels.items()}
        self.labels_rev[-2] = 'Unassigned'

    def save(self, filename: str):
        df = pd.DataFrame(list(self.labels.items()), columns=['token', 'label'])
        df.to_csv(filename, index=False)

    @staticmethod
    def load(filename: str) -> 'Labelling':
        df = pd.read_csv(filename)

        labels: dict[str, int] = {}
        for _, row in df.iterrows():
            labels[str(row['token'])] = int(row['label'])

        return Labelling(labels)


@dataclass
class SplitData:
    index: dict[int, int]
    labels: list[int]
    texts: list[BatchEncoding]

    @staticmethod
    def from_df(df: pd.DataFrame, labels: Labelling, label_threshold: int) -> 'SplitData':
        index = {e: i for i, e in enumerate(df.index.tolist())}
        labels = [-2 if label is None else labels.labels[label] for label in df[OUT_COLUMN]]
        labels = [-1 if label >= label_threshold else label for label in labels]
        texts = [tokenizer(text, padding='max_length', max_length=512, truncation=True,
                           return_tensors='pt') for text in df[IN_JOINED_COLUMN]]
        return SplitData(index, labels, texts)

    def __init__(self, index: dict[int, int], labels: list[int], texts: list[BatchEncoding]):
        self.index = index
        self.labels = labels
        self.texts = texts

    def __len__(self) -> int:
        return len(self.texts)

    def only_issue(self, issue_id: int) -> Optional['SplitData']:
        if issue_id not in self.index:
            return None

        i = self.index[issue_id]
        label = self.labels[i]
        text = self.texts[i]

        return SplitData({issue_id: 0}, [label], [text])


def df_validation_split(df: pd.DataFrame) -> tuple[pd.DataFrame, pd.DataFrame]:
    df_train, df_val = np.split(df, [int((1 - VALIDATION_PERC) * len(df))])
    return df_train, df_val


@dataclass
class Datasets:
    train: SplitData
    validation: SplitData
    test: SplitData
    classifier_label_range: tuple[int, int]
    labelling: Labelling

    def __init__(self, kind: str):
        if kind == 'all':
            df = load_df(f'issues_train_000001_170000.csv')
        elif kind == 'recent':
            df = load_df(f'issues_train_recent_150000_170000.csv')
        else:
            raise ValueError('kind must be one of \'all\' or \'recent\'')

        df_test: pd.DataFrame = load_df(f'issues_test_170001_180000.csv')
        df_train, df_val = df_validation_split(df)

        labels, splits = compute_labels([df_train, df_val, df_test])

        self.labelling = Labelling(labels)
        self.classifier_label_range = (splits[0], splits[1])

        self.train = SplitData.from_df(df_train, self.labelling, splits[1])

        # Remove unknown labels from validation set
        df_val['label_num'] = df_val[OUT_COLUMN].apply(lambda ass: self.labelling.labels[ass])
        df_val = df_val[df_val.label_num < self.classifier_label_range[1]]

        self.validation = SplitData.from_df(df_val, self.labelling, splits[1])

        self.test = SplitData.from_df(df_test, self.labelling, splits[1])
