from dataclasses import dataclass
from typing import Optional

import torch
import tqdm
from torch.utils import data

from .classifier import Classifier
from .dataset import SplitData
from .torch_dataset import Dataset


@dataclass
class PredictionResult:
    top_values: list[float]
    top_indices: list[int]
    truth_idx: Optional[int]

    def __init__(self, top_values: list[float], top_indices: list[int], truth_idx: Optional[int]):
        self.top_values = top_values
        self.top_indices = top_indices
        self.truth_idx = truth_idx

    def __len__(self) -> int:
        return len(self.top_values)


def predict(model: Classifier, test: Dataset, top_n: int, force_cpu: bool) -> list[PredictionResult]:
    batch_size = 16
    test_dataloader = torch.utils.data.DataLoader(test, batch_size=batch_size)

    use_cuda = torch.cuda.is_available() and not force_cpu
    device = torch.device("cuda" if use_cuda and not force_cpu else "cpu")

    if use_cuda:
        model = model.cuda()

    res = []

    with torch.no_grad():
        for test_x, test_y in tqdm.tqdm(test_dataloader, desc="Test"):
            test_y = test_y.to(device)
            mask = test_x['attention_mask'].to(device)
            input_id = test_x['input_ids'].squeeze(1).to(device)

            output = model(input_id, mask)
            norm_output = torch.softmax(output.logits, dim=1)

            dim_n = norm_output.size(dim=1)
            top_n = min(dim_n if top_n == -1 else top_n, dim_n)

            top = torch.topk(norm_output, top_n, dim=1)

            for i in range(top.values.size(dim=0)):
                res.append(PredictionResult(top.values[i, :].tolist(), top.indices[i, :].tolist(), test_y[i].item()))

    return res


def evaluate(model, test: Dataset, force_cpu: bool, top_k: int = 5):
    test_output = predict(model, test, top_k, force_cpu)

    if len(test_output) > 0:
        top_k = min(top_k, len(test_output[0].top_indices))

    accuracies = [0] * top_k

    for res in test_output:
        for i, index in enumerate(res.top_indices):
            if index == res.truth_idx:
                for j in range(i, len(accuracies)):
                    accuracies[j] += 1

    for i, acc in enumerate(accuracies):
        acc = acc / len(test.texts)
        print(f'Test Accuracy for {i + 1} recommendations: {acc: .4f}')


def predict_top_k(model: Classifier, test_data: SplitData, issue_id: int, top_n: int,
                  force_cpu: bool) -> PredictionResult:
    issue_data = test_data.only_issue(issue_id)

    if issue_data is None:
        raise ValueError("Issue id {0} is not present as an issue in the test set".format(issue_id))

    issue_dataset = Dataset(issue_data)
    result = predict(model, issue_dataset, top_n, force_cpu)

    assert len(result) == 1
    return result[0]
