import torch
from torch import nn
from torch.optim import Adam
from torch.utils import data
from tqdm import tqdm

from .torch_dataset import Dataset


def print_message(epoch_num: int, train_loss: float, train_acc: float, train_ds: Dataset, val_loss: float,
                  val_acc: float, val_ds: Dataset):
    messages = [
        f'Epochs: {epoch_num + 1}',
        f'Train Loss: {train_loss / len(train_ds.texts): .3f}',
        f'Train Accuracy: {train_acc / len(train_ds.texts): .3f}',
        f'Val Loss: {val_loss / len(val_ds.texts): .3f}',
        f'Val Accuracy: {val_acc / len(val_ds.texts): .3f}'
    ]
    print(' | '.join(messages))


def compute_loss_and_acc(label, input_data, device, model) -> tuple[float, float, any]:
    label = label.to(device)
    mask = input_data['attention_mask'].to(device)
    input_id = input_data['input_ids'].squeeze(1).to(device)

    output = model(input_id, attention_mask=mask, labels=label)
    batch_loss = output.loss

    acc = (torch.argmax(output.logits, 1) == label).sum().item()

    return batch_loss.item(), acc, batch_loss


def train(model, train_ds: Dataset, val_ds: Dataset, learning_rate: float, epochs: int, force_cpu: bool):
    batch_size = 16

    train_dataloader = torch.utils.data.DataLoader(train_ds, batch_size=batch_size, shuffle=True)
    val_dataloader = torch.utils.data.DataLoader(val_ds, batch_size=batch_size)

    use_cuda = torch.cuda.is_available() and not force_cpu
    device = torch.device("cuda" if use_cuda and not force_cpu else "cpu")

    if use_cuda:
        model = model.cuda()

    optimizer = Adam(model.parameters(), lr=learning_rate)

    for epoch_num in range(epochs):
        total_acc_train = 0
        total_loss_train = 0

        train_dl = tqdm(train_dataloader, desc=f"Train E{epoch_num + 1}")

        i = 0
        for train_input, train_label in train_dl:

            delta_loss, delta_acc, batch_loss = compute_loss_and_acc(train_label, train_input, device, model)

            total_loss_train += delta_loss
            total_acc_train += delta_acc

            batch_loss.backward()
            optimizer.step()
            model.zero_grad()

            i += train_label.size(dim=0)
            loss_avg = total_loss_train / i
            acc_avg = total_acc_train / i

            train_dl.set_description(f"Train E{epoch_num + 1} loss={loss_avg:.6f} acc={acc_avg:.4f}")

        total_acc_val = 0
        total_loss_val = 0

        val_dl = tqdm(val_dataloader, desc="Val E" + str(epoch_num + 1))

        with torch.no_grad():

            i = 0
            for val_input, val_label in val_dl:
                delta_loss, delta_acc, batch_loss = compute_loss_and_acc(val_label, val_input, device, model)

                total_loss_val += delta_loss
                total_acc_val += delta_acc

                i += val_label.size(dim=0)
                loss_avg = total_loss_val / i
                acc_avg = total_acc_val / i

                val_dl.set_description(f"Val E{epoch_num + 1} loss={loss_avg:.6f} acc={acc_avg:.4f}")

        print_message(epoch_num, total_loss_train, total_acc_train, train_ds, total_loss_val, total_acc_val, val_ds)
