#!/usr/bin/env python

import argparse
import os

os.environ["CUDA_VISIBLE_DEVICES"] = "2"

import torch

from dotenv import load_dotenv

from cleaner.clean import read_issue_obj
from cleaner.dataframe import build_df
from modelimpl.dataset import SplitData, prepare_input, Labelling
from modelimpl.evaluate import predict_top_k
from modelimpl.load import load_model
from scraper.download import download_issue, download_commit_activity

load_dotenv()
TOKEN = os.getenv('GITHUB_TOKEN')


def main(path: str, issue_id: int, force_cpu: bool, top: int):
    stats = download_commit_activity(TOKEN)

    if not path.endswith('.pt'):
        raise ValueError("path should point to a pytorch model file")

    pretrained, model, classes = load_model(path, None, force_cpu, False)

    csv_path = path[:-3] + '.labels.csv'

    labelling = Labelling.load(csv_path)

    if not pretrained:
        raise FileNotFoundError("Trained model is needed to run predict script")

    issue_json = download_issue(issue_id, TOKEN)
    issue_clean = read_issue_obj(issue_json, enable_filter=False)

    if issue_clean is None:
        raise ValueError("Issue does not contain latin characters in title or body, cannot classify")

    issue_df = prepare_input(build_df([issue_clean]))
    issue_dataset = SplitData.from_df(issue_df, labelling, classes)

    res = predict_top_k(model, issue_dataset, issue_id, top, force_cpu)
    for i in range(len(res)):
        value = res.top_values[i]
        idx = res.top_indices[i]
        assignee = labelling.labels_rev[idx]

        print("{0}: '{1}' ({3}) (confidence: {2:.2f}%) ({4} commits authored)"
              .format(i + 1, assignee, value * 100, idx, stats[assignee]))

    if res.truth_idx != -2:
        truth = labelling.labels_rev[res.truth_idx]
        print("Truth: '{0}' ({1}) ({2} commits authored)".format(truth, res.truth_idx, stats[truth]))
    else:
        print("Issue is unassigned on GitHub")


if __name__ == '__main__':
    assert torch is not None  # make sure pytorch is imported and loaded with correct CUDA env variable

    parser = argparse.ArgumentParser(prog='runmodel.py',
                                     description='Model execution script. Downloads a given issue id from the '
                                                 'microsoft/vscode repository, performs the cleaning process and '
                                                 'recommends an assignee using the given model. The script may fail if '
                                                 'the issue title and body do not contain any latin characters.')
    parser.add_argument('modelfile', type=str, help="Path to the pickled pytorch model to classify the issue with")
    parser.add_argument('issue_id', type=int, help="The microsoft/vscode GitHub issue id to classify")
    parser.add_argument('-t', '--top', type=int, default=5, help="Number of recommendations to output")
    parser.add_argument('-c', '--force-cpu', action='store_true', help="disables CUDA support. Useful when debugging")

    args = parser.parse_args()
    main(args.modelfile, args.issue_id, args.force_cpu, args.top)
