import http.client
import io
import json
import os
from collections import defaultdict
from tarfile import TarFile, TarInfo
from urllib.parse import urlencode

import tqdm
from dotenv import load_dotenv

load_dotenv()

GITHUB_JSON_MIME: str = 'application/vnd.github+json'
TOKEN_PREFIX: str = 'Bearer '
GITHUB_API: str = 'api.github.com'
REPO_ENDPOINT: str = '/repos/microsoft/vscode'
ISSUE_ENDPOINT: str = REPO_ENDPOINT + '/issues'
JSON_OUT_DIR: str = os.path.join(os.path.dirname(__file__), '..', 'out')


def get_res_body_or_fail(connection: http.client.HTTPSConnection) -> any:
    res = connection.getresponse()
    if res.status // 100 != 2:
        raise IOError("Response status from Github is " + str(res.status) + "\n" + res.read().decode())

    res_text = res.read().decode()
    return json.loads(res_text)


def download_page(page: int, token: str, tar_file: TarFile) -> bool:
    if page < 1:
        raise ValueError("page must be >= 1")

    per_page = 100
    total_estimate = 200000  # rough estimate of the number of issues to download

    connection = http.client.HTTPSConnection(GITHUB_API)
    headers = {'Accept': GITHUB_JSON_MIME,
               'User-Agent': 'usi-msde-2023-soft-analytics-bug-triaging-g2',
               'Authorization': TOKEN_PREFIX + token,
               'X-GitHub-Api-Version': '2022-11-28'}
    query = {'state': 'closed', 'page': str(page), 'per_page': str(per_page)}
    connection.request('GET', ISSUE_ENDPOINT + '?' + urlencode(query),
                       headers=headers)
    res_json = get_res_body_or_fail(connection)

    for issue in tqdm.tqdm(res_json, desc="Downloading page " + str(page), initial=(page - 1) * per_page,
                           total=total_estimate):
        issue_id: int = issue['id']
        filename: str = os.path.join(JSON_OUT_DIR, str(issue_id) + '.json')
        contents: str = json.dumps(issue)

        tar_info = TarInfo(name=filename)
        tar_info.size = len(contents)

        file_object = io.BytesIO(contents.encode())
        tar_file.addfile(tar_info, fileobj=file_object)

    return len(res_json) > 0


def download_issue(issue_id: int, token: str) -> dict[str, any]:
    connection = http.client.HTTPSConnection(GITHUB_API)
    headers = {'Accept': GITHUB_JSON_MIME,
               'User-Agent': 'usi-msde-2023-soft-analytics-bug-triaging-g2',
               'Authorization': TOKEN_PREFIX + token,
               'X-GitHub-Api-Version': '2022-11-28'}
    connection.request('GET', ISSUE_ENDPOINT + '/' + str(issue_id), headers=headers)
    return get_res_body_or_fail(connection)


def download_commit_activity(token: str) -> defaultdict[str, int]:
    connection = http.client.HTTPSConnection(GITHUB_API)
    headers = {'Accept': GITHUB_JSON_MIME,
               'User-Agent': 'usi-msde-2023-soft-analytics-bug-triaging-g2',
               'Authorization': TOKEN_PREFIX + token,
               'X-GitHub-Api-Version': '2022-11-28'}
    connection.request('GET', REPO_ENDPOINT + '/stats/contributors', headers=headers)
    res_obj = get_res_body_or_fail(connection)
    ret: defaultdict[str, int] = defaultdict(int)

    for obj in res_obj:
        ret[str(obj["author"]["login"])] = int(obj["total"])

    return ret
