#!/usr/bin/env python

import argparse
import os

from modelimpl.dataset import Datasets

os.environ["CUDA_VISIBLE_DEVICES"] = "2"

import torch

from modelimpl.evaluate import evaluate
from modelimpl.load import load_model, get_model_path
from modelimpl.torch_dataset import Dataset
from modelimpl.torch_train import train

OUT_DIR = os.path.join(os.path.dirname(__file__), '', '..', 'out', 'model')


def main(dataset_kind: str, epochs: int, learning_rate: float, force_cpu: bool, force_retrain: bool):
    datasets = Datasets(dataset_kind)
    print('Training for dataset kind:', dataset_kind)
    print('Train set instance size: ', len(datasets.train))
    print('Validation set instance size: ', len(datasets.validation))
    print('Test set instance size: ', len(datasets.test))

    path = get_model_path(dataset_kind, epochs, learning_rate, 'pt')
    pretrained, model, _ = load_model(path, datasets.classifier_label_range, force_cpu, force_retrain)

    if pretrained:
        print('Using already trained model')
    else:
        if not os.path.isdir(OUT_DIR):
            os.makedirs(OUT_DIR)

        print('Training model then saving in ' + path)
        train(model, Dataset(datasets.train), Dataset(datasets.validation), learning_rate, epochs, force_cpu)

        torch.save(model.state_dict(), path)
        datasets.labelling.save(get_model_path(dataset_kind, epochs, learning_rate, 'labels.csv'))
        with open(get_model_path(dataset_kind, epochs, learning_rate, 'label_range.txt'), "w") as f:
            f.writelines([str(x) + "\n" for x in datasets.classifier_label_range])

    evaluate(model, Dataset(datasets.test), force_cpu)


if __name__ == '__main__':
    assert torch is not None  # make sure pytorch is imported and loaded with correct CUDA env variable

    parser = argparse.ArgumentParser(prog='trainmodel.py',
                                     description='Training and evaluation script. The script will train and save the '
                                                 'obtained model and then perform test set evaluation. If the given '
                                                 'parameters match with a model that was already saved, the script '
                                                 'only runs the evaluation procedure.')
    parser.add_argument('dataset', choices=['all', 'recent'], type=str, help="The dataset to train with")
    parser.add_argument('epochs', type=int, help="Number of epochs of the training process")
    parser.add_argument('-r', '--learning-rate', type=float, default=1e-6,
                        help="The learning rate fed in the Adam optimizer")
    parser.add_argument('-c', '--force-cpu', action='store_true',
                        help="disables CUDA support. Useful when debugging")
    parser.add_argument('-f', '--force-retraining', action='store_true',
                        help="forces training of a new model even if a matching model is already found within the "
                             "saved models")

    args = parser.parse_args()
    main(args.dataset, args.epochs, args.learning_rate, args.force_cpu, args.force_retraining)
